//
// ip_tunnel.cc    : IP connector Main File
// author          : Clarence Chung
//
// Copyright (C) 2000-2005 by the University of Southern California
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

/*
 *
 * Copyright (c) 2003 The Regents of the University of California.  All 
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Neither the name of the University nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "ip_tunnel.hh"
#include <unistd.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h>
#include <linux/ip.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


NRSimpleAttributeFactory < int >DstAddrAttr (DST_IP_ADDR_KEY,
                                             NRAttribute::INT32_TYPE);
NRSimpleAttributeFactory < void *>IpPktAttr (IP_PKT_KEY,
                                             NRAttribute::BLOB_TYPE);
NRSimpleAttributeFactory < int >IpPktSizeAttr (IP_PKT_SZ_KEY,
                                               NRAttribute::INT32_TYPE);


void
IpConnectorReceive::recv (NRAttrVec * data, NR::handle my_handle)
{
  app_->recv (data, my_handle);
}

void
IpConnectorApp::recv (NRAttrVec * data, NR::handle my_handle)
{
  NRSimpleAttribute < int >*dstaddrAttr = NULL;
  NRSimpleAttribute < void *>*ippktAttr = NULL;
  NRSimpleAttribute < int >*ippktsizeAttr = NULL;
  char *ipbuf;
  int ret, i;
  unsigned int srcIP, dstIP;

  dstaddrAttr = DstAddrAttr.find (data);
  ippktAttr = IpPktAttr.find (data);
  ippktsizeAttr = IpPktSizeAttr.find (data);

  if (dstaddrAttr && ippktAttr && ippktsizeAttr)
    {
      ipbuf = (char *) ippktAttr->getVal ();

      srcIP = ntohl (((struct iphdr *) ipbuf)->saddr);
      dstIP = ntohl (((struct iphdr *) ipbuf)->daddr);
      DiffPrint (DEBUG_DETAILS,
                 "Recv'd %d bytes diffusion IP pkt(srcIP: %0X dstIP: %0X) from filter core:",
                 ippktsizeAttr->getVal (), srcIP, dstIP);

      for (i = 0; i < ippktsizeAttr->getVal (); i++)
        {
          if (i % 16 == 0)
            DiffPrint (DEBUG_DETAILS, "\n\t");
          if (i % 8 == 0)
            DiffPrint (DEBUG_DETAILS, " ");
          DiffPrint (DEBUG_DETAILS, "%02X ", (unsigned char) ipbuf[i]);
        }
      DiffPrint (DEBUG_DETAILS, "\n");

      ret = write (tunfd, ipbuf, ippktsizeAttr->getVal ());
      if (ret < 0)
        {
          DiffPrint (DEBUG_DETAILS, "Error sending to TUN interface.\n");
        }
      else
        {
          DiffPrint (DEBUG_DETAILS,
                     "Sending IP pkt of size %d to TUN interface.\n\n",
                     ippktsizeAttr->getVal ());
        }

    }
  else
    {
      DiffPrint (DEBUG_ALWAYS, "Received a BAD packet !\n");
      PrintAttrs (data);
    }
}

handle IpConnectorApp::setupSubscription ()
{
  NRAttrVec
    attrs;

  attrs.push_back (NRClassAttr.
                   make (NRAttribute::IS, NRAttribute::INTEREST_CLASS));
  attrs.push_back (NRAlgorithmAttr.
                   make (NRAttribute::IS,
                         NRAttribute::ONE_PHASE_PULL_ALGORITHM));
//  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS, NRAttribute::TWO_PHASE_PULL_ALGORITHM));
  attrs.push_back (DstAddrAttr.make (NRAttribute::EQ, myIP));

  handle
    h = dr_->subscribe (&attrs, mr_);

  ClearAttrs (&attrs);

  return h;
}

handle IpConnectorApp::setupPublication ()
{
  NRAttrVec
    attrs;

  attrs.push_back (NRClassAttr.
                   make (NRAttribute::IS, NRAttribute::DATA_CLASS));
  attrs.push_back (NRAlgorithmAttr.
                   make (NRAttribute::IS,
                         NRAttribute::ONE_PHASE_PULL_ALGORITHM));
//  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS, NRAttribute::TWO_PHASE_PULL_ALGORITHM));

  handle
    h = dr_->publish (&attrs);

  ClearAttrs (&attrs);

  return h;
}

void
IpConnectorApp::run ()
{
  char *buf = getenv ("node_addr");
  fd_set fds;
  char ipbuf[1600], dev[100];
  int retval, i, ippkt_sz;
  unsigned long int srcIP, dstIP;

  if (buf == NULL)
    {
      DiffPrint (DEBUG_ALWAYS, "Error: node_addr not set.\n");
      exit (1);
    }


  if (input_addr == NULL)
    {
      myIP = DEFAULT_IP;
      myNetMask = DEFAULT_SUBNET_MASK;
    }

  node_id_ = atoi (buf);
  myIP |= node_id_;


  strcpy (dev, "tun%d");

  /* Create tun interface */
  tunfd = tun_alloc (dev);

  // Setup publication and subscription
  subHandle_ = setupSubscription ();
  pubHandle_ = setupPublication ();


  // Main thread will send ping probes
  while (1)
    {
      FD_ZERO (&fds);
      FD_SET (tunfd, &fds);

      select (tunfd + 1, &fds, NULL, NULL, NULL);

      if (FD_ISSET (tunfd, &fds))
        {

          ippkt_sz = read (tunfd, ipbuf, sizeof (ipbuf));

          srcIP = ntohl (((struct iphdr *) ipbuf)->saddr);
          dstIP = ntohl (((struct iphdr *) ipbuf)->daddr);
          DiffPrint (DEBUG_DETAILS,
                     "Read %i bytes IP pkt(srcIP: %0X, dst IP: %0X) from tun interface:",
                     ippkt_sz, srcIP, dstIP);

          for (i = 0; i < ippkt_sz; i++)
            {
              if (i % 16 == 0)
                DiffPrint (DEBUG_DETAILS, "\n\t");
              if (i % 8 == 0)
                DiffPrint (DEBUG_DETAILS, " ");
              DiffPrint (DEBUG_DETAILS, "%02X ", (unsigned char) ipbuf[i]);
            }
          DiffPrint (DEBUG_DETAILS, "\n");

          dstaddrAttr_ = DstAddrAttr.make (NRAttribute::IS, dstIP);
          data_attr_.push_back (dstaddrAttr_);

          ippktAttr_ =
            IpPktAttr.make (NRAttribute::IS, (void *) ipbuf, ippkt_sz);
          data_attr_.push_back (ippktAttr_);

          ippktsizeAttr_ = IpPktSizeAttr.make (NRAttribute::IS, ippkt_sz);
          data_attr_.push_back (ippktsizeAttr_);

          retval = dr_->send (pubHandle_, &data_attr_);

          if (retval < 0)
            DiffPrint (DEBUG_ALWAYS, "Error sending to filter core.\n");
          else
            DiffPrint (DEBUG_DETAILS,
                       "Sending IP pkt of size %d to filter core\n\n",
                       ippkt_sz);

          ClearAttrs (&data_attr_);
        }
    }                           // end while 
}

IpConnectorApp::IpConnectorApp (int argc, char **argv)
{
  mr_ = new IpConnectorReceive (this);

  input_addr = NULL;
  myMTU = DEFAULT_MTU;

  parseCommandLine (argc, argv);
  dr_ = NR::createNR (diffusion_port_);
}

int
main (int argc, char **argv)
{
  IpConnectorApp *app;

  app = new IpConnectorApp (argc, argv);
  app->run ();

  return 0;
}

// The following function was adapted from the ip_connector emstar
// module. Both University of Southern California and The University of
// California copyright notices apply (see notices on top of file)
int
IpConnectorApp::tun_alloc (char *dev)
{
  struct ifreq ifr;
  int fd, err, n;
  char cmdbuf[200], ipbuf[20], subnetbuf[20];
  long owner = geteuid ();

  if ((fd = open ("/dev/net/tun", O_RDWR)) < 0)
    {
      DiffPrint (DEBUG_ALWAYS, "Cannont open /dev/net/tun\n"
                 "make sure you're running as root\n");
      exit (1);
    }

  memset (&ifr, 0, sizeof (ifr));

  /* Flags: IFF_TUN   - TUN device (no Ethernet headers)
   *        IFF_TAP   - TAP device
   *
   *        IFF_NO_PI - Do not provide packet information
   */
  ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
  if (*dev)
    strncpy (ifr.ifr_name, dev, IFNAMSIZ);

  if ((err = ioctl (fd, TUNSETIFF, (void *) &ifr)) < 0)
    {
      close (fd);
      return err;
    }
  strcpy (dev, ifr.ifr_name);

  // set ownership---ioctl
  if (ioctl (fd, TUNSETOWNER, owner) < 0)
    {
      DiffPrint (DEBUG_ALWAYS, "ioctl TUNSETOWNER Error\n");
      exit (1);
    }

  if (ioctl (fd, TUNSETNOCSUM, 1) < 0)
    {
      DiffPrint (DEBUG_ALWAYS, "ioctl TUNSETNOCSUM Error\n");
      exit (1);
    }

  sprintf (ipbuf, "%d.%d.%d.%d",
           ((unsigned int) myIP >> 24),
           ((unsigned int) myIP >> 16) & 0x000000FF,
           ((unsigned int) myIP >> 8) & 0x000000FF,
           ((unsigned int) myIP & 0x000000FF));
  sprintf (subnetbuf, "%d.%d.%d.%d", ((unsigned int) myNetMask >> 24),
           ((unsigned int) myNetMask >> 16) & 0x000000FF,
           ((unsigned int) myNetMask >> 8) & 0x000000FF,
           ((unsigned int) myNetMask & 0x000000FF));

  sprintf (cmdbuf,
           "/sbin/ifconfig %s %s netmask %s -arp -allmulti -broadcast mtu %d up\n",
           ifr.ifr_name, ipbuf, subnetbuf, myMTU);

  n = system (cmdbuf);

  if (n == 0)
    {
      DiffPrint (DEBUG_DETAILS,
                 "Successufully configured TUN interface '%s' with command: %s",
                 ifr.ifr_name, cmdbuf);
    }
  else
    {
      DiffPrint (DEBUG_ALWAYS, "Couldn't create interface; "
                 "make sure you're running as root\n");
      exit (1);
    }

  return fd;
}

void
IpConnectorApp::usage (char *s)
{
  DiffPrint (DEBUG_ALWAYS,
             "Usage: %s [-d debug] [-p port] [-a IP address prefix] [-m MTU] [-f file] [-h]\n\n",
             s);
  DiffPrint (DEBUG_ALWAYS, "\t-d - Sets debug level (0-10)\n");
  DiffPrint (DEBUG_ALWAYS, "\t-p - Uses port 'port' to talk to diffusion\n");
  DiffPrint (DEBUG_ALWAYS,
             "\t-a - Specifies IP address prefix (ex: 10.0.0.0/24)\n");
  DiffPrint (DEBUG_ALWAYS, "\t-h - Prints this information\n");
  DiffPrint (DEBUG_ALWAYS, "\n");
  exit (0);
}

void
IpConnectorApp::parseCommandLine (int argc, char **argv)
{
  u_int16_t diff_port = DEFAULT_DIFFUSION_PORT;
  int debug_level;
  int opt;
  char *addrbuf;
  int bits;
  unsigned long int addr, mask;

  opterr = 0;

  while (1)
    {
      opt = getopt (argc, argv, "a:m:hd:p:");
      switch (opt)
        {

        case 'p':

          diff_port = (u_int16_t) atoi (optarg);
          if ((diff_port < 1024) || (diff_port >= 65535))
            {
              DiffPrint (DEBUG_ALWAYS,
                         "Error: Diffusion port must be between 1024 and 65535 !\n");
              exit (-1);
            }

          break;

        case 'h':

          usage (argv[0]);

          break;

        case 'd':

          debug_level = atoi (optarg);

          if (debug_level < 1 || debug_level > 10)
            {
              DiffPrint (DEBUG_ALWAYS,
                         "Error: Debug level outside range or missing !\n");
              usage (argv[0]);
            }

          global_debug_level = debug_level;

          break;

        case 'm':

          myMTU = atoi (optarg);

          if (myMTU <= 0)
            {
              DiffPrint (DEBUG_ALWAYS, "Error: Invalid MTU %d!\n", myMTU);
              usage (argv[0]);
            }

          break;

        case 'a':

          if (!strncasecmp (optarg, "-", 1))
            {
              DiffPrint (DEBUG_ALWAYS, "Error: Parameter missing !\n");
              usage (argv[0]);
            }

          input_addr = strdup (optarg);

          addrbuf = strchr (input_addr, '/');

          if (addrbuf == NULL)
            {
              DiffPrint (DEBUG_ALWAYS,
                         "Error: invalid CIDR address format '%s'.\n",
                         input_addr);
              usage (argv[0]);
            }

          *addrbuf++ = 0;
          bits = atoi (addrbuf);

          if ((addr = ntohl (inet_addr (input_addr))) == INADDR_NONE ||
              (bits < 0 || bits > 32))
            {
              DiffPrint (DEBUG_ALWAYS,
                         "Error: invalid CIDR address format '%s'.\n",
                         input_addr);
              usage (argv[0]);
            }

          mask = 0xFFFFFFFF;
          bits = 32 - bits;
          while (bits--)
            mask = mask << 1;

          myIP = addr & mask;
          myNetMask = mask;

          break;

        case '?':

          DiffPrint (DEBUG_ALWAYS,
                     "Error: %c isn't a valid option or its parameter is missing !\n",
                     optopt);
          usage (argv[0]);

          break;

        case ':':

          DiffPrint (DEBUG_ALWAYS, "Parameter missing !\n");
          usage (argv[0]);

          break;

        }

      if (opt == -1)
        break;
    }

  diffusion_port_ = diff_port;
}
