//
// rbp_sender.cc : Rbp Server Main File
// author        : Fred Stann
//

#include "rbp_sender.hh"
#include <unistd.h>

void RbpSenderReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  app_->recv(data, my_handle);
}

void RbpSenderApp::recv(NRAttrVec *data, NR::handle my_handle)
{
  NRSimpleAttribute<int> *nrclass = NULL;

  nrclass = NRClassAttr.find(data);

  if (!nrclass){
    DiffPrint(DEBUG_ALWAYS, "RbpSender::Received a BAD packet !\n");
    return;
  }

  switch (nrclass->getVal()){

  case NRAttribute::INTEREST_CLASS:

    DiffPrint(DEBUG_ALWAYS, "RbpSender::Received an Interest message !\n");
    num_subscriptions_++;
    break;

  case NRAttribute::DISINTEREST_CLASS:

    DiffPrint(DEBUG_ALWAYS, "RbpSender::Received a Disinterest message !\n");
    num_subscriptions_--;
    break;

  default:

    DiffPrint(DEBUG_ALWAYS, "RbpSender::Received an unknown message (%d)!\n", nrclass->getVal());
    break;

  }
}

// Setup what kind of data we have available

handle RbpSenderApp::setupPublication()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::DATA_CLASS));
  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS, NRAttribute::ONE_PHASE_PULL_ALGORITHM));
  // Use two-phase-pull semantics; publications are local
  attrs.push_back(NRScopeAttr.make(NRAttribute::IS, NRAttribute::NODE_LOCAL_SCOPE));
  attrs.push_back(RbpTargetAttr.make(NRAttribute::IS, "Rbp"));

  handle h = dr_->publish(&attrs);

  ClearAttrs(&attrs);
  return h;
}

// Tell diffusion what kind of interests we are interested in

handle RbpSenderApp::setupSubscription()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::EQ, NRAttribute::INTEREST_CLASS));
  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS, NRAttribute::ONE_PHASE_PULL_ALGORITHM));
  attrs.push_back(NRScopeAttr.make(NRAttribute::IS, NRAttribute::NODE_LOCAL_SCOPE));
  attrs.push_back(RbpTargetAttr.make(NRAttribute::IS, "Rbp"));

  handle h = dr_->subscribe(&attrs, mr_);

  ClearAttrs(&attrs);
  return h;
}

void RbpSenderApp::run()
{
  //  int retval;

  // Tell diffusion what kind of interests we latch
  subHandle_ = setupSubscription();
  // Setup publication
  pubHandle_ = setupPublication();

  // Create counter attribute
  counterAttr_ = RbpCounterAttr.make(NRAttribute::IS, last_seq_sent_);
  data_attr_.push_back(counterAttr_);

  // Create the Source ID attribute
  sourceIDAttr_ = RbpSrcIdAttr.make(NRAttribute::IS, source_id_);
  data_attr_.push_back(sourceIDAttr_);

  // Main thread will send ping probes
  while(1){
    sleep(SEND_DATA_INTERVAL);

    // Send data packet if we have active subscriptions
    if (num_subscriptions_ > 0){

      // Send data probe
      //DiffPrint(DEBUG_ALWAYS, "RbpSender::Sending Data %d\n", last_seq_sent_);
      //DiffPrint(DEBUG_ALWAYS, "RbpSender::Waking but not Sending Data\n");
      //retval = dr_->send(pubHandle_, &data_attr_);

      // Update counter
      last_seq_sent_++;
      counterAttr_->setVal(last_seq_sent_);
    }
    else
      DiffPrint(DEBUG_ALWAYS, "RbpSender::no subscribers yet!\n");
  }
}

RbpSenderApp::RbpSenderApp(int argc, char **argv)
{
  last_seq_sent_ = 0;
  num_subscriptions_ = 0;

  mr_ = new RbpSenderReceive(this);

  parseCommandLine(argc, argv);
  dr_ = NR::createNR(diffusion_port_);
}

int main(int argc, char **argv)
{
  RbpSenderApp *app;

  app = new RbpSenderApp(argc, argv);
  app->run();

  return 0;
}
