//
// rbp_filter.cc  : Reliable Broadcast Propagation Filter
// author         : Fred Stann
//

#include "rbp_filter.hh"

RbpFilter *app;

char *generic_msg_types[] = {"INTEREST", "POSITIVE REINFORCEMENT", "NEGATIVE REINFORCEMENT",
		"DATA", "EXPLORATORY DATA", "PUSH EXPLORATORY DATA", "CONTROL",
                "REDIRECT", "LOCAL REPAIR REQUEST", "LOCAL REPAIR REPLY", "STATUS"};

// Whenever the core sends the Rbp Filter a message, this gets called.
void RbpFilterReceive::recv(Message *msg, handle h)
{
  app_->recv(msg, h);
}

void RbpFilter::recv(Message *msg, handle h)
{
  if (h == pre_filter_handle_){
    // Process the message handed to us by the core.
    // If true is returned we forward the message to the next filter. Otherwise it dies here.
    if (preProcessMessage(msg))
      ((DiffusionRouting *)dr_)->sendMessage(msg, pre_filter_handle_, RBP_PRE_FILTER_PRIORITY);
    return;
  }
  else if (h == post_filter_handle_){
      postProcessMessage(msg);
      ((DiffusionRouting *)dr_)->sendMessage(msg, post_filter_handle_, RBP_POST_FILTER_PRIORITY);
  }
  else if (h == status_filter_handle_){
      processStatusMessage(msg);
  }
  else
    DiffPrint(DEBUG_ALWAYS, "RbpFilter(%d)::Error - recv received message for handle %d !\n",
      node_id_, (int)h);
}

// The pre-filter handles broadcasts (INTEREST & EXPLORATORY).
bool RbpFilter::preProcessMessage(Message *msg)
{
  timeval cur_time;
  GetTime (&cur_time);
  int cur_secs = cur_time.tv_sec;
  BroadcastEntry *bcast_entry;
  BroadcastMap::iterator bcast_iterator;
  union LlToKey key;
  NRSimpleAttribute<void *> *rbp_bcast_ack_attr = NULL;
  NRSimpleAttribute<int> *rbp_imp_link_attr = NULL;
  NRSimpleAttribute<void *> *rbp_reinf_ack_attr = NULL;

  // Print some debug strings
#if 1
  if ((msg->new_message_) && (msg->msg_type_ != STATUS)){
    if (msg->last_hop_ != LOCALHOST_ADDR)
      DiffPrint(DEBUG_DETAILS, "preProcessMessage Received a new %s message from node %d\n",
        generic_msg_types[msg->msg_type_], msg->last_hop_);
    else
      DiffPrint(DEBUG_DETAILS, "preProcessMessage Received a new %s message from local agent\n",
         generic_msg_types[msg->msg_type_]);
  }
  else if (msg->msg_type_ != STATUS){
    if (msg->last_hop_ != LOCALHOST_ADDR)
      DiffPrint(DEBUG_DETAILS, "preProcessMessage Received an old %s message from node %d\n",
        generic_msg_types[msg->msg_type_], msg->last_hop_);
    else
      DiffPrint(DEBUG_DETAILS, "preProcessMessage Received an old %s message from local agent\n",
        generic_msg_types[msg->msg_type_]);
  }
#endif

  // Ignore NODE_LOCAL_SCOPE interests
  if (msg->msg_type_ == INTEREST){
    NRSimpleAttribute<int> *nrscope = NULL;
    nrscope = NRScopeAttr.find(msg->msg_attr_vec_);
    if (nrscope->getVal() == NRAttribute::NODE_LOCAL_SCOPE){
      // DiffPrint(DEBUG_DETAILS, "   INTEREST is NODE_LOCAL_SCOPE, so ignore!\n");
      return true;
    }
  }

  // ****** Section 1 Handle broadcast messages

  // We need to log any Resource Discovery message that is flooded.
  if ( (msg->msg_type_ == INTEREST) || (msg->msg_type_ == EXPLORATORY_DATA) || 
       (msg->msg_type_ == PUSH_EXPLORATORY_DATA) ){
    key.int_val_[0] = msg->pkt_num_;
    key.int_val_[1] = msg->rdm_id_;

    // If we don't have this message in our log, we need to add it,
    // and we set a timer.
    bcast_iterator = broadcast_log_.find(key.ll_val_);
    if (bcast_iterator == broadcast_log_.end()){
      TimerCallback *bcast_timer;
      DiffPrint(DEBUG_DETAILS, "   New BroadcastMap entry; Key = %llx\n", key.ll_val_);
      bcast_entry = new BroadcastEntry(msg);
      bcast_entry->secs_ = cur_secs;
      bcast_entry->pkt_num_ = msg->pkt_num_;
      bcast_entry->rdm_id_ = msg->rdm_id_;
      // Update the neighbor_acks_ list unless we generated the message.
      if (msg->last_hop_ != LOCALHOST_ADDR){
        bcast_entry->addNeighborAcked(msg->last_hop_);
        DiffPrint(DEBUG_DETAILS, "   Add neighbor %d to neighbor_acks_\n", msg->last_hop_);
      }
      else
        DiffPrint(DEBUG_DETAILS, "   Don't add neighbor because LOCALHOST\n");

      // Insert the BroadcastEntry into the broadcast_log_
      broadcast_log_.insert(BroadcastMap::value_type(key.ll_val_, bcast_entry));

      // Now set a RelBroadcastTimer
      bcast_timer = new RelBroadcastTimer(this, key.ll_val_);
      bcast_entry->rel_bcast_timer_handle_ = ((DiffusionRouting *)dr_)->addTimer (BCAST_TIMER_DELAY + (int) ((BCAST_TIMER_JITTER * (GetRand() * 1.0 / RAND_MAX) - (BCAST_TIMER_JITTER / 2))), bcast_timer);

      // Update the new_bcast_cnt_ from this neighbor; This is for important link calculation
      if (msg->last_hop_ != LOCALHOST_ADDR){
        NeighborList::iterator neighbor_itr;
        neighbor_itr = neighbors_.begin();
        while (neighbor_itr != neighbors_.end()){
          if (neighbor_itr->neighbor_id_ == msg->last_hop_){
            // Neighbor found, update new broadcast count
            neighbor_itr->new_bcast_cnt_++;
            break;
          }
          // Not found yet, move to next entry
          neighbor_itr++;
        }
      }

      // Update the global new_broadcast_count_ needed for  important link calculation
      // Start a timer to switch to "normal" mode
      new_bcast_count_++;
      // If this is the first new broadcast start the ImpLinkTimer
      if (imp_link_timer_state_ != IMP_LINK_DONE && new_bcast_count_ == 1 && msg->last_hop_ != LOCALHOST_ADDR && !imp_link_timer_active_){
        TimerCallback *imp_link_timer;
        // Start an important link timer
        DiffPrint(DEBUG_DETAILS, "   Start the important link timer since we got first bcast\n");
        imp_link_timer = new ImpLinkTimer(this);
        imp_link_timer_handle_ = ((DiffusionRouting *)dr_)->addTimer(605000 + (int) ((BCAST_TIMER_JITTER * (GetRand() * 1.0 / RAND_MAX) - (BCAST_TIMER_JITTER / 2))), imp_link_timer);
        imp_link_timer_active_ = true;
      }

      // Log the new arrival of a new Broadcast Message, since we never saw it before
      addLogEntry(NEW_BCAST_ARRIVAL, key.ll_val_, msg->last_hop_);
    }
    // We have this entry, update the neighborAcked list for an "implicit" ACK
    else {
      bcast_entry = (*bcast_iterator).second;
      DiffPrint(DEBUG_DETAILS, "   We have this BroadcastMap entry; Key = %llx\n", key.ll_val_);

      // Now update the Acked Map.
      if(!bcast_entry->neighborAcked(msg->last_hop_)){
        bcast_entry->addNeighborAcked(msg->last_hop_);
        DiffPrint(DEBUG_DETAILS, "   Adding neighbor %d to Acked Map\n", msg->last_hop_);
        // Log that we got an initial ACK (implicit) from this neighbor
        addLogEntry(FIRST_IMPL_ACK, key.ll_val_, msg->last_hop_);
      }
      // If this neighbor already sent this broadcast/ACK, we may send an explicit Broadcast ACK!
      else{
        RbpAckBlob *ack_blob;
        NRAttribute *bcast_ack_attr;
        NRAttrVec attrs;
        Message *bcast_ack_msg;

        // If we never explicitly ACKED this guy - send him one.
        if(!bcast_entry->weExplAckedNeighbor(msg->last_hop_)){
            ack_blob = new RbpAckBlob(bcast_entry->pkt_num_, bcast_entry->rdm_id_);
            DiffPrint(DEBUG_DETAILS, "   this neighbor already sent this msg, send ACK to node %d.\n",
	        msg->last_hop_);
            bcast_ack_attr = RbpBcastAckAttr.make(NRAttribute::IS, (void *)ack_blob, sizeof(RbpAckBlob));
            attrs.push_back(bcast_ack_attr);
            // Code to send an ack message to last_hop_
            bcast_ack_msg = new Message(DIFFUSION_VERSION, DATA, 0, 0, attrs.size(),
              pkt_count_, rdm_id_, msg->last_hop_, LOCALHOST_ADDR);
            pkt_count_++;
            bcast_ack_msg->msg_attr_vec_ = CopyAttrs(&attrs);
            ((DiffusionRouting *)dr_)->sendMessage(bcast_ack_msg, pre_filter_handle_,
              ROUTING_COMPLETED_PRIORITY);
            ClearAttrs(&attrs);
            delete bcast_ack_msg;
	    // Mark that we sent an explicit ACK
            bcast_entry->setExplAckedNeighbor(msg->last_hop_);
            // Log that we sent an Explicit ACK
            addLogEntry(SENT_EXPL_ACK, key.ll_val_, msg->last_hop_);
	}
        // Log that we got a redundant ACK (Implicit)
        addLogEntry(RED_IMPL_ACK, key.ll_val_, msg->last_hop_);
      }
    }
  }

  // ****** Section 2 Handle "explicit" broadcast ACKs

  // If this is a Broadcast ACK we process it and return false.
  rbp_bcast_ack_attr = RbpBcastAckAttr.find(msg->msg_attr_vec_);
  if (rbp_bcast_ack_attr){
    RbpAckBlob *rbp_ack_blob;
    rbp_ack_blob = (RbpAckBlob *) rbp_bcast_ack_attr->getVal();
    key.int_val_[0] = rbp_ack_blob->pkt_num_;
    key.int_val_[1] = rbp_ack_blob->rdm_id_;
    DiffPrint(DEBUG_DETAILS, "   got an RBP ACK for msg ID %llx\n", key.ll_val_);

    // Find the BroadcastEntry and update the acked list!
    bcast_iterator = broadcast_log_.find(key.ll_val_);
    if (bcast_iterator != broadcast_log_.end()){
      DiffPrint(DEBUG_DETAILS, "   Found BroadcastEntry matching ACK!\n");
      bcast_entry = (*bcast_iterator).second;
      if(!bcast_entry->neighborAcked(msg->last_hop_)){
        DiffPrint(DEBUG_DETAILS, "   Can't find neighbor %d in Acked Map\n", msg->last_hop_);
        bcast_entry->addNeighborAcked(msg->last_hop_);
        DiffPrint(DEBUG_DETAILS, "   Adding neighbor %d to Acked Map\n", msg->last_hop_);
        // Indicate we got an initial ACK (explicit) from this neighbor
        addLogEntry(FIRST_EXPL_ACK, key.ll_val_, msg->last_hop_);
      }
      else{
        DiffPrint(DEBUG_DETAILS, "   neighbor %d already Acked\n", msg->last_hop_);
        // Indicate we got a redundant ACK (explicit)
        addLogEntry(RED_EXPL_ACK, key.ll_val_, msg->last_hop_);
      }
    }
    else
      DiffPrint(DEBUG_ALWAYS, "   Can't find BroadcastEntry for ACK\n");

    return false;
  }
 
  // ****** Section 3 handle important link notification or ACK

  // If this is an Important Link notification, we process it and return false.
  rbp_imp_link_attr = RbpCtlAttr.find(msg->msg_attr_vec_);
  if (rbp_imp_link_attr){
    int rbp_ctl_type;
    rbp_ctl_type = rbp_imp_link_attr->getVal();
    if (rbp_ctl_type == RBP_IMP_LINK){
      NRAttrVec attrs;
      Message *ctl_msg;
      DiffPrint(DEBUG_DETAILS, "   neighbor %d sez he's an important link!!!\n", msg->last_hop_);
      // See if this is redundant; only add a new important neighbor notification
      ImpNeighborList::iterator imp_neighbor_iter = important_neighbors_.begin();
      while (imp_neighbor_iter != important_neighbors_.end()){
        if(msg->last_hop_ == (*imp_neighbor_iter))
          break;
        imp_neighbor_iter++;
      }
      if (imp_neighbor_iter == important_neighbors_.end()){
        DiffPrint(DEBUG_DETAILS, "   add neighbor %d to important_neighbors!!!\n", msg->last_hop_);
        important_neighbors_.push_back(msg->last_hop_);
      }
      else
        DiffPrint(DEBUG_DETAILS, "   we already have neighbor %d in important_neighbors!!!\n", msg->last_hop_);

      // Now send an important link ack
      attrs.push_back(RbpCtlAttr.make(NRAttribute::IS, RBP_IMP_LINK_ACK));
      ctl_msg = new Message(DIFFUSION_VERSION, DATA, 0,0, attrs.size(),
                pkt_count_, rdm_id_, msg->last_hop_, LOCALHOST_ADDR);
      ctl_msg->msg_attr_vec_ = CopyAttrs(&attrs);
      pkt_count_++;
      // Important neighbor notification, send an ACK to this neighbor!
      DiffPrint(DEBUG_DETAILS, "preProcessMessage: send an RBP_IMP_LINK_ACK message to %d\n", msg->last_hop_);
      ((DiffusionRouting *)dr_)->sendMessage(ctl_msg, pre_filter_handle_,
            ROUTING_COMPLETED_PRIORITY);
      ClearAttrs(&attrs);
      delete ctl_msg;
      addLogEntry(SENT_IMP_LINK_ACK, 0, msg->last_hop_);
    }
    else if (rbp_ctl_type == RBP_IMP_LINK_ACK){
      DiffPrint(DEBUG_DETAILS, "   neighbor %d has acked important link notification!!!\n", msg->last_hop_);
      if (msg->last_hop_ == imp_link_neighbor_)
        imp_link_neighbor_acked_ = true;
      else
        DiffPrint(DEBUG_DETAILS, "   neighbor %d acked, but we were looking for %d!!!\n",
          msg->last_hop_, imp_link_neighbor_);
      addLogEntry(GOT_IMP_LINK_ACK, 0, msg->last_hop_);
    }
    return false;
  }

  // ****** Section 4 handle Positive Reinforcement ACKs

  // If this is a Positive Reinforcement ACK we process it and return false.
  rbp_reinf_ack_attr = RbpReinfAckAttr.find(msg->msg_attr_vec_);
  if (rbp_reinf_ack_attr){
    PosReinfEntry *pos_reinf_entry;
    PosReinfMap::iterator pos_reinf_iterator;
    RbpAckBlob *rbp_ack_blob;
    rbp_ack_blob = (RbpAckBlob *) rbp_reinf_ack_attr->getVal();
    key.int_val_[0] = rbp_ack_blob->pkt_num_;
    key.int_val_[1] = rbp_ack_blob->rdm_id_;
    DiffPrint(DEBUG_DETAILS, "   got a Reinforcement ACK for msg ID %llx\n", key.ll_val_);
    // Update the acked_ variable in the reinf_acks_ entry
    // Find the entry
    pos_reinf_iterator = reinf_acks_.find(key.ll_val_);
    if (pos_reinf_iterator == reinf_acks_.end()){
      DiffPrint(DEBUG_ALWAYS, "   Can't find cached positive reinforcement, kill timer!\n");
    }
    else{
      pos_reinf_entry = (*pos_reinf_iterator).second;
      pos_reinf_entry->acked_ = true;
    }
    return false;
  }

  // ****** Section 5 handle Positive Reinforcements

  // If this is a Positive Reinforcement we need to send an ACK in guaranteed mode.
  if (msg->msg_type_ == POSITIVE_REINFORCEMENT){
    NRSimpleAttribute<void *> *reinforcement_attr = NULL;
    ReinforcementBlob *reinforcement_blob;
    RbpAckBlob *ack_blob;
    NRAttribute *reinf_ack_attr;
    NRAttrVec attrs;
    Message *reinf_ack_msg;

    reinforcement_attr = ReinforcementAttr.find(msg->msg_attr_vec_);
    if (!reinforcement_attr){
      DiffPrint(DEBUG_ALWAYS, "   Can't find a reinforcement blob!\n");
      return true;
    }

    // Extract Reinforcement Blob
    reinforcement_blob = (ReinforcementBlob *) reinforcement_attr->getVal();

    key.int_val_[0] = reinforcement_blob->pkt_num_;
    key.int_val_[1] = reinforcement_blob->rdm_id_;
    DiffPrint(DEBUG_DETAILS, "   Reinforcement is for msg ID %llx\n", key.ll_val_);

    // We must prepare and send a Reinf ACK
    ack_blob = new RbpAckBlob(key.int_val_[0], key.int_val_[1]);
    reinf_ack_attr = RbpReinfAckAttr.make(NRAttribute::IS, (void *)ack_blob, sizeof(RbpAckBlob));
    attrs.push_back(reinf_ack_attr);
    // Code to send an ack message to last_hop_
    DiffPrint(DEBUG_DETAILS, "   Sending reinforcement ACK to node %d\n", msg->last_hop_);
    reinf_ack_msg = new Message(DIFFUSION_VERSION, DATA, 0, 0, attrs.size(),
      pkt_count_, rdm_id_, msg->last_hop_, LOCALHOST_ADDR);
    pkt_count_++;
    reinf_ack_msg->msg_attr_vec_ = CopyAttrs(&attrs);
    ((DiffusionRouting *)dr_)->sendMessage(reinf_ack_msg, pre_filter_handle_,
      ROUTING_COMPLETED_PRIORITY);
    ClearAttrs(&attrs);
    delete reinf_ack_msg;
  }
  return true;
}

void RbpFilter::postProcessMessage(Message *msg){
  PosReinfEntry *reinf_entry;
  TimerCallback *reinf_timer;
  NRSimpleAttribute<void *> *reinforcement_attr = NULL;
  ReinforcementBlob *reinforcement_blob;
  union LlToKey key;
  timeval cur_time;
  GetTime (&cur_time);
  // int cur_secs = cur_time.tv_sec;
  // Print some debug strings
#if 0
  if (msg->new_message_){
    if (msg->last_hop_ != LOCALHOST_ADDR)
      DiffPrint(DEBUG_DETAILS, "postProcessMessage Received a new %s message from node %d\n",
        generic_msg_types[msg->msg_type_], msg->last_hop_);
    else
      DiffPrint(DEBUG_DETAILS, "postProcessMessage Received a new %s message from local agent\n",
        generic_msg_types[msg->msg_type_]);
  }
  else{
    if (msg->last_hop_ != LOCALHOST_ADDR)
      DiffPrint(DEBUG_DETAILS, "postProcessMessage Received an old %s message from node %d\n",
        generic_msg_types[msg->msg_type_], msg->last_hop_);
    else
      DiffPrint(DEBUG_DETAILS, "postProcessMessage Received an old %s message from local agent\n",
        generic_msg_types[msg->msg_type_]);
  }
#endif

  // Extract Reinforcement Blob
  reinforcement_attr = ReinforcementAttr.find(msg->msg_attr_vec_);
  if (!reinforcement_attr){
    DiffPrint(DEBUG_ALWAYS, "  can't find ReinforcementAttr!\n");
    return;
  }
  reinforcement_blob = (ReinforcementBlob *) reinforcement_attr->getVal();
  key.int_val_[0] = reinforcement_blob->pkt_num_;
  key.int_val_[1] = reinforcement_blob->rdm_id_;
  DiffPrint(DEBUG_DETAILS, "postProcessMessage sees POS REINFORCEMENT for msg ID %llx, start timer\n", key.ll_val_);
  reinf_entry = new PosReinfEntry(msg);
  reinf_timer = new RelPosReinfTimer(this, key.ll_val_);
  reinf_entry->pos_reinf_timer_handle_ = ((DiffusionRouting *)dr_)->addTimer(2000, reinf_timer);
  // Insert the PosReinfEntry into the reinf_acks_ map
  reinf_acks_.insert(PosReinfMap::value_type(key.ll_val_, reinf_entry));
}

void RbpFilter::processStatusMessage(Message *msg){
  NRSimpleAttribute<int> *link_status_attr;
  NRSimpleAttribute<void *> *link_neighbors_attr;
  NeighborList::iterator neighbor_itr;
  NeighborEntry neighbor_entry;
  int32_t *current_neighbor;
  struct timeval tmv;
  int num_neighbors;

  // Check message type
  if (msg->msg_type_ != STATUS){
    DiffPrint(DEBUG_DETAILS, "processStatusMessage::Status callback received a %s message !\n",
              msg_types[msg->msg_type_]);
    return;
  }

  // Find relevant status attributes
  link_status_attr = LinkStatusAttr.find(msg->msg_attr_vec_);
  link_neighbors_attr = LinkNeighborsAttr.find(msg->msg_attr_vec_);

  // Make sure we got both
  if (!link_status_attr || !link_neighbors_attr){
    DiffPrint(DEBUG_ALWAYS, "processStatusMessage::Attributes missing from status message !\n");
    return;
  }

  // Check link_status_attr
  if (link_status_attr->getVal() != NRAttribute::NEIGHBORS_UPDATE){
    DiffPrint(DEBUG_ALWAYS, "processStatusMessage::Unknown status update !\n");
    return;
  }

  // Get time for reference
  GetTime(&tmv);

  // Process link neighbors update
  num_neighbors = (int) (link_neighbors_attr->getLen() / sizeof(int));
  // DiffPrint(DEBUG_DETAILS, "processStatusMessage::Processing %d neighbors !\n", num_neighbors);

  // Point to first neighbor
  current_neighbor = (int32_t *) link_neighbors_attr->getGenericVal();
  // Update our neighbor list with new neighbors
  for (int i = 0; i < num_neighbors; i++){

    // Try to find neighbor in the neighbor list
    neighbor_itr = neighbors_.begin();

    while (neighbor_itr != neighbors_.end()){
      if (neighbor_itr->neighbor_id_ == *current_neighbor){
        // Neighbor found, update timestamp
        // DiffPrint(DEBUG_DETAILS, "   Neighbor %d already in list\n",
        //         *current_neighbor);
        neighbor_itr->tmv_ = tmv;
        break;
      }

      // Not found yet, move to next entry
      neighbor_itr++;
    }

    // Check if we need to add this neighbor
    if (neighbor_itr == neighbors_.end()){
      // Yes, this is a new neighbor !
      neighbor_entry.neighbor_id_ = *current_neighbor;
      neighbor_entry.tmv_ = tmv;
      neighbor_entry.new_bcast_cnt_ = 0;
      neighbors_.push_back(neighbor_entry);

      DiffPrint(DEBUG_DETAILS, "   Neighbor %d added to list\n", *current_neighbor);
    }

    // Move to next neighbor in the link update
    current_neighbor++;
  }
  // Step 2: Check if there were any neighbors NOT updated !
  neighbor_itr = neighbors_.begin();

  while (neighbor_itr != neighbors_.end()){
    if ((tmv.tv_sec != neighbor_itr->tmv_.tv_sec) ||
        (tmv.tv_usec != neighbor_itr->tmv_.tv_usec)){

      ImpNeighborList::iterator imp_neighbor_iter = important_neighbors_.begin();
      while (imp_neighbor_iter != important_neighbors_.end()){
        if(neighbor_itr->neighbor_id_ == (*imp_neighbor_iter)){
	  // Delete this neighbor from the important neighbor list
          DiffPrint(DEBUG_DETAILS, "   Important Neighbor %d removed!\n", *imp_neighbor_iter);
	  important_neighbors_.erase(imp_neighbor_iter);
          break;
	}
        imp_neighbor_iter++;
      }
      // Delete neighbor from neighbor list
      DiffPrint(DEBUG_DETAILS, "   Neighbor %d removed from list !\n",
                neighbor_itr->neighbor_id_);
      neighbor_itr = neighbors_.erase(neighbor_itr);
    }
    else{
      // DiffPrint(DEBUG_DETAILS, "   Neighbor %d has the correct timestamp-don't delete!\n",
      //           neighbor_itr->neighbor_id_);
      neighbor_itr++;
    }
  }
  // DiffPrint(DEBUG_DETAILS, "   Num Neighbors = %d!\n", neighbors_.size());
}

void  RbpFilter::setupFilter()
{
  NRAttrVec attrs;

  // This is a dummy attribute for the pre-filter that matches everything
  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::INTEREST_CLASS));
  pre_filter_handle_ = ((DiffusionRouting *)dr_)->addFilter(&attrs, RBP_PRE_FILTER_PRIORITY,
					   filter_callback_);
  ClearAttrs(&attrs);

  // We just want to match on the reinforcement blob for the post-filter
  attrs.push_back(ReinforcementAttr.make(NRAttribute::EQ_ANY, NULL, 0));
  post_filter_handle_ = ((DiffusionRouting *)dr_)->addFilter(&attrs, RBP_POST_FILTER_PRIORITY,
					   filter_callback_);
  ClearAttrs(&attrs);

  // We want to receive the link status notification
  attrs.push_back(LinkStatusAttr.make(NRAttribute::EQ, NRAttribute::NEIGHBORS_UPDATE));
  status_filter_handle_ = ((DiffusionRouting *)dr_)->addFilter(&attrs, RBP_STATUS_FILTER_PRIORITY,
                          filter_callback_);
  ClearAttrs(&attrs);
}

void RbpFilter::run()
{
  // Doesn't do anything
  while (1){
    sleep(1000);
  }
}

int RbpFilter::relBroadcastHandler(int64_t key){
  BroadcastEntry *bcast_entry;
  BroadcastMap::iterator bcast_iterator;
  int num_neighbors, neighbor_acked_count, percent_acked, num_retries;
  ImpNeighborList::iterator imp_neighbor_iter;
  TimerCallback *cleanup_timer;
  timeval cur_time;
  GetTime (&cur_time);
  int cur_secs = cur_time.tv_sec;
  bool resend_message = false;
  bool imp_neigh_retry = false;
  bool send_broadcast = true;

  DiffPrint(DEBUG_DETAILS, "relBroadcastHandler(%llx):: executes at %d secs\n", key, cur_secs);

  // Find the entry
  bcast_iterator = broadcast_log_.find(key);
  if (bcast_iterator == broadcast_log_.end()){
    DiffPrint(DEBUG_ALWAYS, "   relBroadcastHandler Can't find cached old broadcast for id %llx, kill timer!\n", key);
    return -1;
  }
  else
    bcast_entry = (*bcast_iterator).second;

  // Now check the neighbor list for Broadcast ACKS and
  // either rebroadcast and return 0, or destroy entry and return -1
  num_neighbors = neighbors_.size();
  neighbor_acked_count = bcast_entry->getNeighborAckedCount();
  if(num_neighbors != 0)
    percent_acked = (int)(100 * ((float)neighbor_acked_count / (float)num_neighbors));
  else
    percent_acked = 100;
  num_retries = bcast_entry->num_retries_;

  DiffPrint(DEBUG_DETAILS,
    "   num_neighbors = %d, neighbor_acked_count = %d, percent = %d\n", num_neighbors, neighbor_acked_count, percent_acked);

  if(percent_acked < BCAST_THRESHOLD){
    if ((num_neighbors <= 3) && (num_retries < 3)){
      resend_message = true;
      // if (neighbor_acked_count == 1)
      //   send_broadcast = false;
    }
    else if ((num_neighbors <= 6) && (num_retries < 2)){
      resend_message = true;
    }
    else if ((percent_acked < 50) && (num_retries < 1)){
      resend_message = true;
    }
  }

  if (  !resend_message && !important_neighbors_.empty() && num_retries < 4 ){
    imp_neighbor_iter = important_neighbors_.begin();
    while (imp_neighbor_iter != important_neighbors_.end()){
      if(!bcast_entry->neighborAcked(*imp_neighbor_iter)){
        DiffPrint(DEBUG_DETAILS, 
          "   Handler sees an unAcked important neighbor %d\n", *imp_neighbor_iter);
        imp_neigh_retry = true;
        resend_message = true;
        break;
      }
      imp_neighbor_iter++;
    }
  }

  if((percent_acked < BCAST_THRESHOLD)&&(num_retries < 4)){
      resend_message = true;
  }

  // At this point the resend_message and imp_neigh_retry variables are set.
  // Now we rebroadcast or reunicast the message, depending on the variables.

  // If we are done trying, mark this BroadcastEntry for deletion and kill this timeout.
  if(!resend_message){
    DiffPrint(DEBUG_DETAILS, 
      "   sees %d of %d neighbors acked (%d percent); done!\n",
       neighbor_acked_count, num_neighbors, percent_acked);
    DiffPrint(DEBUG_DETAILS, 
      "   Handler sees required neighbors Acked, mark for deletion & mark time\n");
    bcast_entry->delete_entry_ = true;
    bcast_entry->secs_ = cur_secs;
    if (cleanup_timer_active_ == false){
      // Set a CleanupTimer
      DiffPrint(DEBUG_DETAILS, "   Start the cleanup timer since its not active\n");
      cleanup_timer = new CleanupTimer(this);
      cleanup_timer_handle_ = ((DiffusionRouting *)dr_)->addTimer(60000, cleanup_timer);
      cleanup_timer_active_ = true;
    }
    // Indicate we are done with this Broadcast
    addLogEntry(BCAST_DONE, key, 0);
    return -1;
  }

  // We are resending this message, either via broadcast or unicast.
  
  int32_t neighbor_id;
  NeighborList::iterator neighbor_iterator;
  Message *recast_msg;

  DiffPrint(DEBUG_DETAILS, "   sees %d of %d neighbors acked (%d percent); num_retries = %d - resend!\n",
       neighbor_acked_count, num_neighbors, percent_acked, num_retries);
  bcast_entry->num_retries_++;
  recast_msg = CopyMessage(bcast_entry->message_);

#if 0
  DiffPrint(DEBUG_DETAILS, "  CopyMessage results version_ %d msg_type_ %d source_port_ %d data_len_ %d num_attr_ %d \n",
      recast_msg->version_, recast_msg->msg_type_, recast_msg->source_port_, recast_msg->data_len_, recast_msg->num_attr_);
  DiffPrint(DEBUG_DETAILS, "    pkt_num_ %x rdm_id_ %x next_hop_ %d last_hop_ %d new_message_ %d next_port_ %d\n",
      recast_msg->pkt_num_, recast_msg->rdm_id_, recast_msg->next_hop_, recast_msg->last_hop_, recast_msg->new_message_, recast_msg->next_port_);
  PrintAttrs(recast_msg->msg_attr_vec_);
#endif
  if (send_broadcast){
    recast_msg->next_hop_ = BROADCAST_ADDR;
    recast_msg->last_hop_ = LOCALHOST_ADDR;
    ((DiffusionRouting *)dr_)->sendMessage(recast_msg, pre_filter_handle_, ROUTING_COMPLETED_PRIORITY);
    // Indicate Rebroadcast
    if (imp_neigh_retry)
      addLogEntry(SPEC_REBROADCAST, key, BROADCAST_ADDR);
    else
      addLogEntry(REBROADCAST, key, BROADCAST_ADDR);
  }
  // We are unicasting!!
  else{
    // We are unicasting only to important neighbors
    if(imp_neigh_retry){
      imp_neighbor_iter = important_neighbors_.begin();
      while (imp_neighbor_iter != important_neighbors_.end()){
        if(!bcast_entry->neighborAcked(*imp_neighbor_iter)){
          recast_msg->next_hop_ = *imp_neighbor_iter;
          recast_msg->last_hop_ = LOCALHOST_ADDR;
          ((DiffusionRouting *)dr_)->sendMessage(recast_msg, pre_filter_handle_, ROUTING_COMPLETED_PRIORITY);
          DiffPrint(DEBUG_DETAILS, 
            "   Handler sees unAcked important neighbor %d, reunicast message\n", *imp_neighbor_iter);
          addLogEntry(SPEC_REUNICAST, key, *imp_neighbor_iter);
        }
        imp_neighbor_iter++;
      }
    }
    // We are unicasting to all unacked neighbors
    else{
      // For each unacked neighbor resend a unicast message
      for(neighbor_iterator = neighbors_.begin(); neighbor_iterator != neighbors_.end();
        ++neighbor_iterator){
        neighbor_id = neighbor_iterator->neighbor_id_;
        if(!bcast_entry->neighborAcked(neighbor_id)){
          recast_msg->next_hop_ = neighbor_id;
          recast_msg->last_hop_ = LOCALHOST_ADDR;
          ((DiffusionRouting *)dr_)->sendMessage(recast_msg, pre_filter_handle_,
             ROUTING_COMPLETED_PRIORITY);
          // log Re-unicast
          addLogEntry(REUNICAST, key, neighbor_id);
        }
      }
    }
  } //Unicasting

  delete recast_msg;
  return ( BCAST_TIMER_DELAY + (int) ((BCAST_TIMER_JITTER * 
    (GetRand() * 1.0 / RAND_MAX) - (BCAST_TIMER_JITTER / 2))) );
}

int RbpFilter::relPosReinfHandler(int64_t key){
  int retval;
  PosReinfEntry *pos_reinf_entry;
  PosReinfMap::iterator pos_reinf_iterator;
  timeval cur_time;
  GetTime (&cur_time);
  int cur_secs = cur_time.tv_sec;

  DiffPrint(DEBUG_DETAILS, "relPosReinfHandler(%llx):: executes at %d secs\n", key, cur_secs);

  // Find the reinf_acks_ entry for the key
  pos_reinf_iterator = reinf_acks_.find(key);
  if (pos_reinf_iterator == reinf_acks_.end()){
    DiffPrint(DEBUG_ALWAYS, "  Can't find cached positive reinforcement, kill timer!\n");
    retval = -1;
  }
  else{
    pos_reinf_entry = (*pos_reinf_iterator).second;
    if (pos_reinf_entry->acked_ == true){
      DiffPrint(DEBUG_DETAILS, "relPosReinfHandler::entry (%llx) has been acked; kill timer\n", key);
      delete pos_reinf_entry;
      reinf_acks_.erase(pos_reinf_iterator);
      retval = -1;
    }
    else{
      Message *reinf_msg;
      DiffPrint(DEBUG_DETAILS, "relPosReinfHandler::entry (%llx) has not been acked; resend Pos Reinf\n", key);
      // Resend Positive Reinforcement
      reinf_msg = CopyMessage(pos_reinf_entry->message_);
      ((DiffusionRouting *)dr_)->sendMessage(reinf_msg, post_filter_handle_,
        RBP_POST_FILTER_PRIORITY);
      delete reinf_msg;
      retval = 0;
    }
  }
  return retval;
}

int RbpFilter::cleanupHandler(){
  BroadcastEntry *bcast_entry;
  BroadcastMap::iterator bcast_iterator;
  union LlToKey key;
  timeval cur_time;
  GetTime (&cur_time);
  int cur_secs = cur_time.tv_sec;

  DiffPrint(DEBUG_DETAILS, "cleanupHandler(%d):: executes at %d secs\n", node_id_, cur_secs);

  // First dump the log if it's big.
  if(log_.size() > 50)
    printLog();


  // Now we delete all BroadcastEntrys marked for deletion some time ago.
  for (bcast_iterator = broadcast_log_.begin(); bcast_iterator != broadcast_log_.end();){
    bcast_entry = (*bcast_iterator).second;
    if (bcast_entry->delete_entry_ && (cur_secs - bcast_entry->secs_ > 360)){
      key.int_val_[0] = bcast_entry->pkt_num_;
      key.int_val_[1] = bcast_entry->rdm_id_;
      DiffPrint(DEBUG_DETAILS, "  Deleting BroadcastMap entry; Key = %llx\n", key.ll_val_);
      delete bcast_entry;
      broadcast_log_.erase(bcast_iterator++);
    }
    else
      ++bcast_iterator;
  }
  if (broadcast_log_.empty()){
    DiffPrint(DEBUG_DETAILS, "  kill the cleanup timer since nothing is in the broadcast map\n");
    cleanup_timer_active_ = false;
    return -1;
  }
  else
    return 0;
}

int RbpFilter::impLinkHandler(){
  timeval cur_time;
  GetTime (&cur_time);
  int cur_secs = cur_time.tv_sec;
  NeighborList::iterator neighbor_itr;
  NRAttrVec attrs;
  Message *ctl_msg;
  int bcast_count = 0;
  float importance;

  // Check if we need to do important link recognition
  if (imp_link_timer_state_ == IMP_LINK_COLLECTING_DATA){
    DiffPrint(DEBUG_DETAILS, "impLinkHandler(%d):: executes at %d secs in COLLECTING_DATA state\n", node_id_, cur_secs);
    DiffPrint(DEBUG_DETAILS, "impLinkHandler(%d):: sees neighbor list of size %d\n", neighbors_.size());
    neighbor_itr = neighbors_.begin();
    while (neighbor_itr != neighbors_.end()){
      bcast_count += neighbor_itr->new_bcast_cnt_;
      neighbor_itr++;
    }
    if (bcast_count == 0)
      bcast_count = 1;
    neighbor_itr = neighbors_.begin();
    while (neighbor_itr != neighbors_.end()){
      importance = (float)neighbor_itr->new_bcast_cnt_ / (float)bcast_count;
      DiffPrint(DEBUG_DETAILS, "impLinkHandler(%d):: sees neighbor %d has new_bcase_cnt_ of %d importance %f\n",
        node_id_, neighbor_itr->neighbor_id_, neighbor_itr->new_bcast_cnt_, importance);
      if ( importance >= .70 ){
        // Send a notification to transmitting neighbor
        attrs.push_back(RbpCtlAttr.make(NRAttribute::IS, RBP_IMP_LINK));
        ctl_msg = new Message(DIFFUSION_VERSION, DATA, 0,0, attrs.size(),
                pkt_count_, rdm_id_, neighbor_itr->neighbor_id_, LOCALHOST_ADDR);
        ctl_msg->msg_attr_vec_ = CopyAttrs(&attrs);
        pkt_count_++;
        // Important neighbor found, send a message to this neighbor!
        imp_link_neighbor_ = neighbor_itr->neighbor_id_;
        imp_link_neighbor_acked_ = false;
	imp_link_retries_ = 0;
        imp_link_timer_state_ = IMP_LINK_WAITING_ACK;
        DiffPrint(DEBUG_DETAILS, "impLinkHandler: send an RBP_IMP_LINK message to %d\n", neighbor_itr->neighbor_id_);
        ((DiffusionRouting *)dr_)->sendMessage(ctl_msg, pre_filter_handle_,
              ROUTING_COMPLETED_PRIORITY);
        ClearAttrs(&attrs);
        delete ctl_msg;
        addLogEntry(SENT_IMP_LINK_MSG, 0, neighbor_itr->neighbor_id_);
        // We can only have 1 neighbor with > .80 so break from while loop
        break;
      }
      //  move to next entry
      neighbor_itr++;
    } // while loop
    if (imp_link_timer_state_ == IMP_LINK_WAITING_ACK)
      // wait 5 seconds for ack with jitter
      return ( 5000 + (int) ((BCAST_TIMER_JITTER * 
          (GetRand() * 1.0 / RAND_MAX) - (BCAST_TIMER_JITTER / 2))) );
    else{
      // We didn't recognize any important link, so just kill the timer
      imp_link_timer_state_ = IMP_LINK_DONE;
      addLogEntry(DONE_IMP_LINK_REC, 0, 0);
      imp_link_timer_active_ = false;  // For now don't collect more data
      return -1; 
    }
  }
  else if (imp_link_timer_state_ == IMP_LINK_WAITING_ACK){
    DiffPrint(DEBUG_DETAILS, "impLinkHandler(%d):: executes at %d secs in WAITING_ACK state\n", node_id_, cur_secs);
    // If we didn't get an ack resend important link message
    if ((imp_link_neighbor_acked_ == false)&&(imp_link_retries_ < 4)){
      attrs.push_back(RbpCtlAttr.make(NRAttribute::IS, RBP_IMP_LINK));
      ctl_msg = new Message(DIFFUSION_VERSION, DATA, 0,0, attrs.size(),
              pkt_count_, rdm_id_, imp_link_neighbor_, LOCALHOST_ADDR);
      ctl_msg->msg_attr_vec_ = CopyAttrs(&attrs);
      pkt_count_++;
      imp_link_retries_++;
      // Important neighbor found, send a message to this neighbor!
      DiffPrint(DEBUG_DETAILS, "impLinkHandler: sends another RBP_IMP_LINK message to %d\n", imp_link_neighbor_);
      addLogEntry(SENT_IMP_LINK_MSG, 0, imp_link_neighbor_);
      ((DiffusionRouting *)dr_)->sendMessage(ctl_msg, pre_filter_handle_,
            ROUTING_COMPLETED_PRIORITY);
      ClearAttrs(&attrs);
      delete ctl_msg;
      // wait 5 seconds for ack with jitter
      return ( 5000 + (int) ((BCAST_TIMER_JITTER * 
          (GetRand() * 1.0 / RAND_MAX) - (BCAST_TIMER_JITTER / 2))) );
    }
    else{
      DiffPrint(DEBUG_DETAILS, "impLinkHandler: sees important neighbor %d acked or we give up\n", imp_link_neighbor_);
      imp_link_timer_state_ = IMP_LINK_DONE;
      addLogEntry(DONE_IMP_LINK_REC, 0, 0);
      // We clean the accumulated new_broadcast counts.
      neighbor_itr = neighbors_.begin();
      while (neighbor_itr != neighbors_.end()){
        neighbor_itr->new_bcast_cnt_ = 0;
        neighbor_itr++;
      }
      new_bcast_count_ = 0;
      imp_link_timer_active_ = false;  // For now don't collect more data
      return -1; 
    }
  }
  else{
    DiffPrint(DEBUG_DETAILS, "impLinkHandler(%d):: executes at %d secs in unknown state\n", node_id_, cur_secs);
    imp_link_timer_active_ = false;
    return -1; 
  }
}

RbpFilter::RbpFilter(int argc, char **argv)
{
  // Create Diffusion Routing class
  parseCommandLine(argc, argv);
  dr_ = NR::createNR(diffusion_port_);

  filter_callback_ = new RbpFilterReceive(this);
  pkt_count_ = rand();
  rdm_id_ = rand();
  cleanup_timer_active_ = false;
  imp_link_timer_active_ = false;
  imp_link_timer_state_ = IMP_LINK_DONE;
  new_bcast_count_ = 0;
  stats_ = true;
  log_file_ = false;
  log_mtx_ = new pthread_mutex_t;
  pthread_mutex_init(log_mtx_, NULL);

  // Set up the filter
#ifndef USE_EMSIM
  char *node_id_env;
  node_id_env = getenv("node_addr");
  if (node_id_env != NULL)
      node_id_ = atoi(node_id_env);
  DiffPrint(DEBUG_ALWAYS, "rbp::constructor gets node id %d from environment var\n",node_id_);
#else
  char *sim_id_env;
  sim_id_env = getenv("SIM_ID");
  if (sim_id_env !=NULL)
      node_id_ = atoi(sim_id_env);
  DiffPrint(DEBUG_ALWAYS, "rbp::constructor gets node id %d from simulator\n",node_id_);
#endif
  setupFilter();

  DiffPrint(DEBUG_DETAILS, "rbp:: pre-filter subscribed to *, received handle %d\n",
	    (int)pre_filter_handle_);
  DiffPrint(DEBUG_DETAILS, "rbp:: filter initialized !\n");
}

void signal_handler(int p)
{
  app->timeToStop();
  exit(0);
}

void RbpFilter::timeToStop()
{
  if (stats_)
      printLog();
}

void RbpFilter::printLog()
{
  char out_filename[100];
  FILE *outfile = NULL;
  LogList::iterator log_itr;

  sprintf(out_filename, "/tmp/logs/rbp-%d.log", node_id_);

  if(!log_file_){
    outfile = fopen(out_filename, "w");
    log_file_ = true;
    fprintf(outfile, "RBP Filter log\n");
    fprintf(outfile, "---------------\n\n");
    fprintf(outfile, "Node id : %d\n", node_id_);
  }
  else{
    outfile = fopen(out_filename, "a");
  }

  if (outfile == NULL){
    DiffPrint(DEBUG_ALWAYS,
              "Diffusion Error: Cannot create %s\n", out_filename);
    return;
  }

  pthread_mutex_lock(log_mtx_);
  log_itr = log_.begin();

  while (log_itr != log_.end()){
    fprintf(outfile, "secs %d  ", log_itr->secs_);
    switch (log_itr->type_){
    case NEW_BCAST_ARRIVAL:
      fprintf(outfile, "NEW_BROADCAST_ARRIVED   ");
      break;
    case FIRST_IMPL_ACK:
      fprintf(outfile, "FIRST_IMPLICIT_ACK      ");
      break;
    case RED_IMPL_ACK:
      fprintf(outfile, "REDUNDANT_IMPLICIT_ACK  ");
      break;
    case SENT_EXPL_ACK:
      fprintf(outfile, "SENT_EXPLICIT_ACK       ");
      break;
    case FIRST_EXPL_ACK:
      fprintf(outfile, "INITIAL_EXPLICIT_ACK    ");
      break;
    case RED_EXPL_ACK:
      fprintf(outfile, "REDUNDANT_EXPLICIT_ACK  ");
      break;
    case BCAST_DONE:
      fprintf(outfile, "BROADCAST_DONE          ");
      break;
    case REBROADCAST:
      fprintf(outfile, "REBROADCAST             ");
      break;
    case SPEC_REBROADCAST:
      fprintf(outfile, "SENT_SPEC_REBROADCAST   ");
      break;
    case REUNICAST:
      fprintf(outfile, "REUNICAST               ");
      break;
    case SPEC_REUNICAST:
      fprintf(outfile, "SENT_SPEC_REUNICAST     ");
      break;
    case SENT_IMP_LINK_MSG:
      fprintf(outfile, "SENT_IMPORTANT_LINK_MSG ");
      break;
    case SENT_IMP_LINK_ACK:
      fprintf(outfile, "SENT_IMPORTANT_LINK_ACK ");
      break;
    case GOT_IMP_LINK_ACK:
      fprintf(outfile, "GOT_IMPORTANT_LINK_ACK  ");
      break;
    case DONE_IMP_LINK_REC:
      fprintf(outfile, "DONE_IMPORTANT_LINK_RECOGNITION");
      break;
    }
    fprintf(outfile, " bcastID %x   neighborID %3d   neighbors %2d   ackedCount %2d   retries %d\n", 
        log_itr->id_, log_itr->neighbor_id_, log_itr->neighbor_count_, log_itr->acked_count_, log_itr->num_retries_);

    // Move to next entry
    log_itr++;
  }
  log_.clear();
  pthread_mutex_unlock(log_mtx_);

  fclose(outfile);
}

void RbpFilter::addLogEntry(int type, int64_t key, int32_t neighbor_id){

  LogEntry log_entry;
  timeval cur_time;
  BroadcastEntry *bcast_entry;
  BroadcastMap::iterator bcast_iterator;
  union LlToKey tmp_key;

  GetTime (&cur_time);

  // Find the entry
  bcast_iterator = broadcast_log_.find(key);
  if (bcast_iterator != broadcast_log_.end()){
    bcast_entry = (*bcast_iterator).second;
    log_entry.acked_count_ = bcast_entry->getNeighborAckedCount();
    log_entry.num_retries_ = bcast_entry->num_retries_;
  }
  else{
    log_entry.acked_count_ = 0;
    log_entry.num_retries_ = 0;
  }

  // Add new log entry
  log_entry.secs_ = cur_time.tv_sec;
  log_entry.type_ = type;
  tmp_key.ll_val_ = key;
  log_entry.id_ = tmp_key.int_val_[0];
  log_entry.neighbor_id_ = neighbor_id;
  log_entry.neighbor_count_ = neighbors_.size();
  pthread_mutex_lock(log_mtx_);
  log_.push_back(log_entry);
  pthread_mutex_unlock(log_mtx_);
}


// Reminder retval of 0 = reschedule with same time
//                   -1 = don't reschedule
//              pos int = msec for new time
int RelBroadcastTimer::expire(){
  int retval;
  retval = filter_->relBroadcastHandler(key_);
  if (retval == -1)
    delete this;
  return retval;
}

int RelPosReinfTimer::expire(){
  int retval;
  retval = filter_->relPosReinfHandler(key_);
  if (retval == -1)
    delete this;
  return retval;
}

int CleanupTimer::expire(){
  int retval;
  retval = filter_->cleanupHandler();
  if (retval == -1)
    delete this;
  return retval;
}

int ImpLinkTimer::expire(){
  int retval;
  retval = filter_->impLinkHandler();
  if (retval == -1)
    delete this;
  return retval;
}

int main(int argc, char **argv)
{
  // Initialize and run the Log Filter
  app = new RbpFilter(argc, argv);
  signal(SIGINT, signal_handler);
  app->run();

  return 0;
}
