//
// rbp_filter.hh       : RPB Filter Include File
// author         : Fred Stann
//

#ifndef _RBP_HH_
#define _RBP_HH_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include "diffapp.hh"
#include <signal.h>
#include <list>
#include <map>
#include <pthread.h>

#define BCAST_THRESHOLD 99
#define BCAST_TIMER_DELAY 10000
#define BCAST_TIMER_JITTER 2500

typedef struct _NeighborEntry {
public:
  int32_t neighbor_id_;
  struct timeval tmv_;
  int new_bcast_cnt_;
} NeighborEntry;
typedef list<NeighborEntry> NeighborList;

// Union to convert pkt_num & rdm_id into a 64 bit key
union LlToKey {
  int64_t ll_val_;
  int int_val_[2];
};

typedef map<int32_t, int, less<int32_t> >BcastAckMap;

// Important link definitions
#define IMP_LINK_COLLECTING_DATA 1
#define IMP_LINK_WAITING_ACK 2
#define IMP_LINK_DONE 3
typedef list<int32_t> ImpNeighborList;

class BroadcastEntry {
public:
  BroadcastEntry(Message *message) {
    message_ = CopyMessage(message);
    delete_entry_ = false;
    num_retries_ = 0;
  };
  ~BroadcastEntry() {
    // delete the Attribute
    delete message_;
    neighbor_acks_.clear();
  };
  Message *message_;
  int pkt_num_;
  int rdm_id_;
  int secs_;
  int num_retries_;
  handle rel_bcast_timer_handle_;
  bool delete_entry_;

  BcastAckMap neighbor_acks_;   // Map of "ACKed" indexed by neighbor
  bool neighborAcked(int32_t neighbor_id){
    BcastAckMap::iterator ack_map_itr;
    ack_map_itr = neighbor_acks_.find(neighbor_id);
    if (ack_map_itr != neighbor_acks_.end())
      return true;
    else 
      return false;
  }
  bool weExplAckedNeighbor(int32_t neighbor_id){
    BcastAckMap::iterator ack_map_itr;
    ack_map_itr = neighbor_acks_.find(neighbor_id);
    if (ack_map_itr != neighbor_acks_.end()){
      if((*ack_map_itr).second >= 2)
        return true;
      else
        return false;
    }
    else 
      return false;
  }
  void addNeighborAcked(int32_t neighbor_id){
    neighbor_acks_.insert(BcastAckMap::value_type(neighbor_id, 0));
    return;
  }
  void setExplAckedNeighbor(int32_t neighbor_id){
    BcastAckMap::iterator ack_map_itr;
    ack_map_itr = neighbor_acks_.find(neighbor_id);
    if (ack_map_itr != neighbor_acks_.end())
      (*ack_map_itr).second += 1;
  }
  u_int32_t getNeighborAckedCount(){
    return neighbor_acks_.size();
  }
};
typedef map<int64_t, BroadcastEntry*, less<int64_t> >BroadcastMap;

class PosReinfEntry {
public:
  PosReinfEntry(Message *message) {
    message_ = CopyMessage(message);
    acked_ = false;
  };
  ~PosReinfEntry() {
    // delete the Message
    delete message_;
  };

  Message *message_;
  bool acked_;
  handle pos_reinf_timer_handle_;
};
typedef map<int64_t, PosReinfEntry*, less<int64_t> >PosReinfMap;

#define NEW_BCAST_ARRIVAL 1
#define FIRST_IMPL_ACK    2
#define RED_IMPL_ACK      3
#define SENT_EXPL_ACK     4
#define FIRST_EXPL_ACK    5
#define RED_EXPL_ACK      6
#define BCAST_DONE        7
#define REBROADCAST       8
#define SPEC_REBROADCAST  9
#define REUNICAST         10
#define SPEC_REUNICAST    11
#define SENT_IMP_LINK_MSG 12
#define SENT_IMP_LINK_ACK 13
#define GOT_IMP_LINK_ACK  14
#define DONE_IMP_LINK_REC 15

typedef struct _LogEntry {
public:
  int secs_;
  int type_;
  int32_t id_;
  int32_t neighbor_id_;
  int neighbor_count_;
  int acked_count_;
  int num_retries_;
} LogEntry;
typedef list<LogEntry> LogList;

#define RBP_PRE_FILTER_PRIORITY 205
#define RBP_POST_FILTER_PRIORITY 45
#define RBP_STATUS_FILTER_PRIORITY 135

class RbpFilter;

class RbpFilterReceive : public FilterCallback {
public:
  RbpFilterReceive(RbpFilter *app) : app_(app) {};
  void recv(Message *msg, handle h);
  RbpFilter *app_;
};

class RbpFilter : public DiffApp {
public:
  RbpFilter(int argc, char **argv);
  virtual ~RbpFilter(){};
  void run();
  void recv(Message *msg, handle h);
  int relBroadcastHandler(int64_t key);
  int relPosReinfHandler(int64_t key);
  void timeToStop();
  void printLog();
  int cleanupHandler();
  int impLinkHandler();
private:
  handle pre_filter_handle_;               // pre-filter handle
  handle post_filter_handle_;              // post-filter handle
  handle status_filter_handle_;            // status-filter handle
  RbpFilterReceive *filter_callback_;      // filter callback for received pkts
  void setupFilter();                      // routine that informs filter core
  int pkt_count_;
  int rdm_id_;
  int new_bcast_count_;
  ImpNeighborList important_neighbors_;
  bool imp_link_timer_active_;
  handle imp_link_timer_handle_;
  int imp_link_timer_state_;
  int32_t imp_link_neighbor_;
  bool imp_link_neighbor_acked_;
  int imp_link_retries_;
  bool cleanup_timer_active_;
  handle cleanup_timer_handle_;
  bool stats_;

  // Message Processing
  bool preProcessMessage(Message *msg);
  void postProcessMessage(Message *msg);
  void processStatusMessage(Message *msg);

  // List of all attributes and their counts
  BroadcastMap broadcast_log_;

  // List of neighbors culled from status message
  NeighborList neighbors_;

  // Support for Positive Reinforcement ACKS
  PosReinfMap reinf_acks_;   // Map of reinforcements & ACKs

  // Log
  LogList log_;
  bool log_file_;
  pthread_mutex_t *log_mtx_;
  void addLogEntry(int type, int64_t key, int32_t neighbor_id);
};

class RelBroadcastTimer: public TimerCallback {
public:
  RelBroadcastTimer(RbpFilter *rbp_flt, int64_t key):
    filter_(rbp_flt), key_(key) {};
  ~RelBroadcastTimer() {};
  int expire();
  RbpFilter *filter_;
  int64_t key_;
};

class RelPosReinfTimer: public TimerCallback {
public:
  RelPosReinfTimer(RbpFilter *rbp_flt, int64_t key):
    filter_(rbp_flt), key_(key) {};
  ~RelPosReinfTimer() {};
  int expire();
  RbpFilter *filter_;
  int64_t key_;
};

class CleanupTimer: public TimerCallback {
public:
  CleanupTimer(RbpFilter *rbp_flt):
    filter_(rbp_flt) {};
  ~CleanupTimer() {};
  int expire();
  RbpFilter *filter_;
};

class ImpLinkTimer: public TimerCallback {
public:
  ImpLinkTimer(RbpFilter *rbp_flt):
    filter_(rbp_flt) {};
  ~ImpLinkTimer() {};
  int expire();
  RbpFilter *filter_;
};

class RbpAckBlob {
public:
  RbpAckBlob(int32_t pkt_num, int32_t rdm_id) :
    pkt_num_(pkt_num), rdm_id_(rdm_id) {};

  int32_t pkt_num_;
  int32_t rdm_id_;
};

// We copy the Reinforcement Blob definition from the tpp filter
// because we need to receive and send positive reinforcements.
class ReinforcementBlob {
public:
  ReinforcementBlob(int32_t rdm_id, int32_t pkt_num) :
    rdm_id_(rdm_id), pkt_num_(pkt_num) {};

  int32_t rdm_id_;
  int32_t pkt_num_;
};

#endif // !_RBP_HH_
